// PV112 2017, test for lesson 2
// Libraries needed: GL, GLEW, GLUT, GLM (in development versions!)

#include <iostream>
#include "PV112.h"

#define _USE_MATH_DEFINES
#include <cmath>

#define GLEW_STATIC
#include "GL/glew.h"
#if defined(_WIN32)
#define NOMINMAX // Make Windows.h not define 'min' and 'max' macros
#include <GL/wglew.h> // Include on Windows
#pragma comment(lib, "glew32s.lib")
#else
#include <GL/glxew.h> // Include on Linux and Mac
#define APIENTRY
#endif

#include <GL/freeglut.h>

#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>

using namespace std;
using namespace PV112;

// Current window size
int win_width = 480;
int win_height = 480;

GLuint program;

GLint position_loc;
GLint PVM_matrix_loc;

PV112Geometry my_cube;

float app_time_s = 0.0f;

// Called when the user presses a key
void key_pressed(unsigned char key, int mouseX, int mouseY)
{
    switch (key)
    {
        case 27:		// VK_ESCAPE
            exit(0);
            break;
        case 'l':
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
            glutPostRedisplay();
            break;
        case 'f':
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
            glutPostRedisplay();
            break;
        case 't':
            glutFullScreenToggle();
            break;
    }
}

// Initializes OpenGL stuff
void init()
{
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClearDepth(1.0);
    glEnable(GL_DEPTH_TEST);

    // Create shader program
    program = CreateAndLinkProgram("vertex.glsl", "fragment.glsl");
    if (0 == program)
        WaitForEnterAndExit();

    // Get attribute and uniform locations
    position_loc = glGetAttribLocation(program, "position");
    int tex_loc = glGetAttribLocation(program, "tex");
    PVM_matrix_loc = glGetUniformLocation(program, "PVM_matrix");

    my_cube = CreateCube(position_loc, -1, tex_loc);
}

// Called when the window needs to be rerendered
void render()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glUseProgram(program);

    // Compute the matrics
    glm::mat4 projection_matrix, view_matrix, model_matrix, PVM_matrix;
    projection_matrix = glm::perspectiveRH(glm::radians(45.0f), float(win_width) / float(win_height), 0.1f, 20.0f);
    view_matrix = glm::lookAtRH(glm::vec3(1.0f, 1.5f, 10.0f), glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3(0.0f, 1.0f, 0.0f));
    model_matrix = glm::mat4(1.0f);
    model_matrix = glm::rotate(model_matrix, glm::radians(app_time_s * 100.0f), glm::vec3(1.0f, 0.0f, 0.0f));
    model_matrix = glm::rotate(model_matrix, glm::radians(app_time_s * 120.0f), glm::vec3(0.0f, 1.0f, 0.0f));
    model_matrix = glm::rotate(model_matrix, glm::radians(app_time_s * 140.0f), glm::vec3(0.0f, 0.0f, 1.0f));
    PVM_matrix = projection_matrix * view_matrix * model_matrix;

    // Draw the cube
    glBindVertexArray(my_cube.VAO);
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    DrawGeometry(my_cube);

    glBindVertexArray(0);
    glUseProgram(0);

    // Swaps the front and back buffer (double-buffering)
    glutSwapBuffers();

    // Post rerender, assuming 60 Hz
    glutPostRedisplay();
    app_time_s += 0.01666f;
}

// Called when the window changes its size
void reshape(int width, int height)
{
    win_width = width;
    win_height = height;

    // Set the area into which we render
    glViewport(0, 0, win_width, win_height);
}

// Callback function to be called when we make an error in OpenGL
void APIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const char* message, const void* userParam)
{
    switch (type)			// Report only some errors
    {
        case GL_DEBUG_TYPE_ERROR:
        case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
        case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
            cout << message << endl;		// Put the breakpoint here
            return;
        default:
            return;
    }
}

int main(int argc, char ** argv)
{
    // Initialize GLUT
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);		// Return from the main loop

    // Set OpenGL Context parameters
    glutInitContextVersion(3, 3);				// Use OpenGL 3.3
    glutInitContextProfile(GLUT_CORE_PROFILE);	// Use OpenGL core profile
    glutInitContextFlags(GLUT_DEBUG);			// Use OpenGL debug context

    // Initialize window
    glutInitWindowSize(win_width, win_height);
    glutCreateWindow("PV112 - prepare for cv2");

    // Initialize GLEW
    glewExperimental = GL_TRUE;
    glewInit();

    SetDebugCallback(simple_debug_callback);

    // Initialize out OpenGL stuff
    init();

    // Register callbacks
    glutDisplayFunc(render);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(key_pressed);

    // Run the main loop
    glutMainLoop();

    return 0;
}
